#
# Copyright (c) 2012-2025 Snowflake Computing Inc. All rights reserved.
#

import argparse
import os
from typing import Callable

from snowflake.snowpark_submit.cluster_mode.job_runner import JobRunner


class SparkConnectJobRunner(JobRunner):
    def __init__(
        self,
        args: argparse.Namespace,
        generate_spark_cmd_args: Callable[[argparse.Namespace], list[str]],
    ) -> None:
        super().__init__(
            args,
            generate_spark_cmd_args,
            client_working_dir="/app/",
            temp_stage_mount_dir="/app/client-src/",
            current_dir=os.path.dirname(os.path.abspath(__file__)),
        )

    def _generate_client_container_args(
        self, client_src_zip_file_path: str
    ) -> list[str]:
        args = []
        if client_src_zip_file_path:
            args.extend(["--zip", client_src_zip_file_path])
        args.append(" ".join(self.generate_spark_cmd_args(self.args)))
        return args

    def _get_image_tag(self, env_var_name: str) -> str:
        """
        First, it checks the provided parameter, then checks the environment variable, then defaults to "latest".
        """

        from_param = self.args.snowpark_connect_version
        from_env = os.getenv(env_var_name)

        if from_param == "latest" or (not from_param and not from_env):
            return "latest"
        if not from_param:
            return from_env
        import re

        if re.match(r"^\d+\.\d+\.\d+$", from_param):
            return from_param
        else:
            raise Exception("Incorrect snowpark-connect-version parameter value.")

    def _server_image_path_sys_registry(self) -> str:
        version_tag = self._get_image_tag("SNOWFLAKE_SYSTEM_REGISTRY_SERVER_IMAGE_TAG")
        return f"/snowflake/images/snowflake_images/spark_connect_for_snowpark_server:{version_tag}"

    def _client_image_path_sys_registry(self) -> str:
        version_tag = self._get_image_tag("SNOWFLAKE_SYSTEM_REGISTRY_CLIENT_IMAGE_TAG")
        return f"/snowflake/images/snowflake_images/spark_connect_for_snowpark_client_{self.client_app_language}:{version_tag}"

    def _client_image_name_override(self) -> str:
        return f"snowpark-connect-client-{self.client_app_language}:latest"

    def _server_image_name_override(self) -> str:
        return "snowpark-connect-server:latest"

    def _add_additional_jars_to_classpath(self) -> None:
        # spark-connect-client-jvm_2.12-3.5.3.jar is copied in the docker image.
        self._add_class_paths(
            ["/app/spark_lib/spark-connect-client-jvm_2.12-3.5.3.jar"]
        )

    def _use_system_registry(self) -> bool:
        return os.getenv("SNOWFLAKE_USE_SYSTEM_REGISTRY", "True").lower() == "true"

    def _override_args(self) -> None:
        # For Spark Connect, if customer does not explicitly pass --remote,
        # we always set it to 'sc://localhost:15002'
        if not self.args.remote:
            self.args.remote = "sc://localhost:15002"



#
# Copyright (c) 2012-2025 Snowflake Computing Inc. All rights reserved.
#

import os
from unittest.mock import Mock

import pytest
from airflow.utils.context import Context
from snowflake.operators.snowpark_submit.snowpark_submit_operator import (
    SnowparkSubmitOperator,
    generate_service_name,
)

from snowflake.connector.config_manager import CONFIG_MANAGER

from .conftest import DEFAULT_TEST_CONNECTION_NAME


class TestSnowparkSubmitOperator:
    @classmethod
    def setup_class(cls):
        cls.tests_path = os.path.dirname(os.path.abspath(__file__))
        cls.project_root = os.path.dirname(cls.tests_path)

        cls.test_python_file = os.path.join(cls.tests_path, "resources/pyspark-test.py")
        cls.test_main_file = os.path.join(
            cls.tests_path, "resources/snowpark-submit-pyspark-example/main.py"
        )
        cls.test_jar_file = os.path.join(
            cls.tests_path,
            "resources/snowpark-submit-scala-example/target/original-scala-maven-example-0.1.0.jar",
        )

        snowpark_config = CONFIG_MANAGER["connections"][DEFAULT_TEST_CONNECTION_NAME]

        cls.connection_config = {
            "account": snowpark_config.get("account"),
            "host": snowpark_config.get("host"),
            "user": snowpark_config.get("user"),
            "password": snowpark_config.get("password"),
            "role": snowpark_config.get("role"),
            "warehouse": snowpark_config.get("warehouse"),
            "database": snowpark_config.get("database"),
            "schema": snowpark_config.get("schema"),
            "compute_pool": snowpark_config.get("compute_pool"),
        }

        mock_ti = Mock()
        mock_ti.xcom_push = Mock()
        cls.mock_context = Context()
        cls.mock_context["ti"] = mock_ti

    # ------------- UNIT TESTS -------------

    def test_missing_file_error(self):
        operator = SnowparkSubmitOperator(
            task_id="test_missing_file",
            file="/nonexistent/file.py",
            connections_config=self.connection_config,
        )

        with pytest.raises(FileNotFoundError, match="File not found"):
            operator.execute(self.mock_context)

    def test_unsupported_file_type_error(self):
        import tempfile

        with tempfile.NamedTemporaryFile(suffix=".txt", delete=False) as tf:
            temp_file_path = tf.name

        try:
            operator = SnowparkSubmitOperator(
                task_id="test_unsupported_file",
                file=temp_file_path,
                connections_config=self.connection_config,
            )

            with pytest.raises(
                ValueError,
                match="Unsupported file type: .txt. Only .py and .jar files are supported",
            ):
                operator.execute(self.mock_context)
        finally:
            os.unlink(temp_file_path)

    def test_jar_file_requires_main_class(self):
        import tempfile

        with tempfile.NamedTemporaryFile(suffix=".jar", delete=False) as tf:
            temp_jar_path = tf.name

        try:
            operator = SnowparkSubmitOperator(
                task_id="test_jar_without_main_class",
                file=temp_jar_path,
                connections_config=self.connection_config,
            )

            with pytest.raises(
                ValueError, match="main_class is required for JAR files"
            ):
                operator.execute(self.mock_context)
        finally:
            os.unlink(temp_jar_path)

    def test_connections_config_validation(self):
        incomplete_config = {
            "account": "test_account",
            "host": "test_host.snowflakecomputing.com",
        }

        operator = SnowparkSubmitOperator(
            task_id="test_validation",
            file=self.test_python_file,
            connections_config=incomplete_config,
        )

        with pytest.raises(
            ValueError, match="Missing required Snowflake configuration parameters"
        ):
            operator.execute(self.mock_context)

    def test_service_name_generation(self):
        name1 = generate_service_name()
        name2 = generate_service_name()
        name3 = generate_service_name()

        assert name1.startswith("SNOWPARK_SUBMIT_AIRFLOW_OPERATOR_")
        assert name2.startswith("SNOWPARK_SUBMIT_AIRFLOW_OPERATOR_")
        assert name3.startswith("SNOWPARK_SUBMIT_AIRFLOW_OPERATOR_")

        assert name1 != name2 != name3, "Service names should be unique"
        assert len(name1) == len("SNOWPARK_SUBMIT_AIRFLOW_OPERATOR_") + 27

    def test_wait_for_completion_command_building(self):
        operator1 = SnowparkSubmitOperator(
            task_id="test_cmd_async",
            file=self.test_python_file,
            connections_config=self.connection_config,
        )
        cmd1 = operator1._build_command()
        assert "--wait-for-completion" not in cmd1

        operator2 = SnowparkSubmitOperator(
            task_id="test_cmd_blocking",
            file=self.test_python_file,
            connections_config=self.connection_config,
            wait_for_completion=True,
        )
        cmd2 = operator2._build_command()
        assert "--wait-for-completion" in cmd2

    # ------------- EXECUTION TESTS -------------
    # These tests require SPCS infrastructure
    @pytest.mark.skip(
        reason="Requires SPCS testing infra. Temporarily skipped for to avoid long merge gate time. Can be enabled in the future with more testing resources."
    )
    def test_operator_basic(self):
        operator = SnowparkSubmitOperator(
            task_id="test_cluster_basic",
            file=self.test_python_file,
            connections_config=self.connection_config,
        )

        result = operator.execute(self.mock_context)

        assert result is not None
        assert result["return_code"] == 0
        assert result["service_name"] is not None
        assert result["service_name"].startswith("SNOWPARK_SUBMIT_AIRFLOW_OPERATOR_")
        assert self.mock_context["ti"].xcom_push.call_count == 4
        calls = self.mock_context["ti"].xcom_push.call_args_list
        xcom_data = {call[1]["key"]: call[1]["value"] for call in calls}

        assert xcom_data["service_name"] == result["service_name"]
        assert xcom_data["return_code"] == 0

        stdout_content = result["stdout"]
        service_name = result["service_name"]
        assert (
            f"Job {service_name} has been submitted and running asynchorously in SPCS"
            in stdout_content
        ), f"Job submission to SPCS not confirmed for service {service_name}"

    @pytest.mark.skip(
        reason="Requires SPCS testing infra. Temporarily skipped for to avoid long merge gate time. Can be enabled in the future with more testing resource"
    )
    def test_custom_service_name(self):
        custom_name = "CUSTOM_TEST_SERVICE_12345"
        operator = SnowparkSubmitOperator(
            task_id="test_custom_service_name",
            file=self.test_python_file,
            connections_config=self.connection_config,
            service_name=custom_name,
        )

        result = operator.execute(self.mock_context)

        assert result is not None
        assert result["return_code"] == 0
        assert result["service_name"] == custom_name

        stdout_content = result["stdout"]
        assert (
            f"Job {custom_name} has been submitted and running asynchorously in SPCS"
            in stdout_content
        ), f"Job submission to SPCS not confirmed for custom service {custom_name}"

    @pytest.mark.skip(
        reason="Requires SPCS testing infra. Temporarily skipped for to avoid long merge gate time. Can be enabled in the future with more testing resource"
    )
    def test_properties_file_parameter(self):
        properties_file_path = os.path.join(
            self.tests_path, "resources/test_spark_properties.conf"
        )

        operator = SnowparkSubmitOperator(
            task_id="test_properties_file",
            file=self.test_python_file,
            connections_config=self.connection_config,
            properties_file=properties_file_path,
        )

        cmd = operator._build_command()
        assert "--properties-file" in cmd
        props_idx = cmd.index("--properties-file")
        assert cmd[props_idx + 1] == properties_file_path

        result = operator.execute(self.mock_context)

        assert result is not None
        assert result["return_code"] == 0

        stdout_content = result["stdout"]
        service_name = result["service_name"]
        assert (
            f"Job {service_name} has been submitted and running asynchorously in SPCS"
            in stdout_content
        )

    def test_missing_properties_file_error(self):
        operator = SnowparkSubmitOperator(
            task_id="test_missing_properties_file",
            file=self.test_python_file,
            connections_config=self.connection_config,
            properties_file="/nonexistent/spark-properties.conf",
        )

        with pytest.raises(FileNotFoundError, match="Properties file not found"):
            operator.execute(self.mock_context)

    # ------------- WAIT FOR COMPLETION EXECUTION TESTS -------------
    @pytest.mark.skip(
        reason="Requires SPCS testing infra. Temporarily skipped for to avoid long merge gate time. Can be enabled in the future with more testing resource"
    )
    def test_wait_for_completion_blocking_mode(self):
        operator = SnowparkSubmitOperator(
            task_id="test_blocking_mode",
            file=self.test_python_file,
            connections_config=self.connection_config,
            wait_for_completion=True,
            fail_on_error=True,
        )

        result = operator.execute(self.mock_context)

        assert result is not None
        assert result["return_code"] == 0
        assert result["service_name"] is not None
        assert result["service_name"].startswith("SNOWPARK_SUBMIT_AIRFLOW_OPERATOR_")

        stdout_content = result["stdout"]
        service_name = result["service_name"]

        assert (
            "Workload Status: DONE" in stdout_content
        ), f"Job completion not confirmed in blocking mode for service {service_name}"

    # @pytest.mark.skip(reason="Requires SPCS testing infra. Temporarily skipped for to avoid long merge gate time. Can be enabled in the future with more testing resource")
    def test_wait_for_completion_scala_jar(self):
        operator = SnowparkSubmitOperator(
            task_id="test_blocking_scala_jar",
            file=self.test_jar_file,
            main_class="com.example.SnowparkConnectApp",
            application_args=["TEST_SCALA_OUTPUT_TABLE"],
            connections_config=self.connection_config,
            comment="Scala test execution in blocking mode",
            wait_for_completion=True,
            fail_on_error=True,
        )

        result = operator.execute(self.mock_context)

        assert result is not None
        assert result["return_code"] == 0
        assert result["service_name"] is not None
        assert result["service_name"].startswith("SNOWPARK_SUBMIT_AIRFLOW_OPERATOR_")

        stdout_content = result["stdout"]
        service_name = result["service_name"]

        assert (
            "Workload Status: DONE" in stdout_content
        ), f"Scala JAR job completion not confirmed in blocking mode for service {service_name}"

    @pytest.mark.skip(
        reason="Requires SPCS testing infra. Temporarily skipped for to avoid long merge gate time. Can be enabled in the future with more testing resource"
    )
    def test_operator_with_pyfiles(self, pyspark_utils_zip):
        operator = SnowparkSubmitOperator(
            task_id="test_cluster_pyfiles",
            file=self.test_main_file,
            connections_config=self.connection_config,
            py_files=str(pyspark_utils_zip.resolve()),
            application_args=["test_pyfiles_table"],
            wait_for_completion=True,
        )

        result = operator.execute(self.mock_context)

        assert result is not None
        assert result["return_code"] == 0
        assert result["service_name"] is not None
        assert result["service_name"].startswith("SNOWPARK_SUBMIT_AIRFLOW_OPERATOR_")

        stdout_content = result["stdout"]
        service_name = result["service_name"]

        assert (
            "Workload Status: DONE" in stdout_content
        ), f"Job completion not confirmed in blocking mode for service {service_name}"

    @pytest.mark.skip(
        reason="Requires SPCS testing infra. Temporarily skipped for to avoid long merge gate time. Can be enabled in the future with more testing resource"
    )
    def test_wait_for_completion_job_failure(self):
        failing_python_file = os.path.join(
            self.tests_path, "resources/pyspark-test-failing.py"
        )

        operator = SnowparkSubmitOperator(
            task_id="test_job_failure",
            file=failing_python_file,
            connections_config=self.connection_config,
            wait_for_completion=True,
            fail_on_error=True,
        )

        with pytest.raises(Exception, match="Spark job failed with status: FAILED"):
            operator.execute(self.mock_context)

    @pytest.mark.skip(
        reason="Requires SPCS testing infra. Temporarily skipped for to avoid long merge gate time. Can be enabled in the future with more testing resource"
    )
    def test_wait_for_completion_job_failure_tolerant(self):
        failing_python_file = os.path.join(
            self.tests_path, "resources/pyspark-test-failing.py"
        )

        operator = SnowparkSubmitOperator(
            task_id="test_job_failure_tolerant",
            file=failing_python_file,
            connections_config=self.connection_config,
            wait_for_completion=True,
            fail_on_error=False,  # Shouldnt cause Airflow task to fail
        )

        result = operator.execute(self.mock_context)

        assert result is not None
        assert result["return_code"] == 0
        assert result["service_name"] is not None

        stdout_content = result["stdout"]
        assert (
            "Workload Status: FAILED" in stdout_content
        ), "Expected to see job failure status in blocking mode"



#
# Copyright (c) 2012-2025 Snowflake Computing Inc. All rights reserved.
#

import argparse
import os
from typing import Callable

from snowflake.snowpark_submit.cluster_mode.job_runner import JobRunner


class SparkClassicJobRunner(JobRunner):
    def __init__(
        self,
        args: argparse.Namespace,
        generate_spark_cmd_args: Callable[[argparse.Namespace], list[str]],
    ) -> None:
        super().__init__(
            args,
            generate_spark_cmd_args,
            client_working_dir="/home/spark/",
            temp_stage_mount_dir="/home/spark/client-src/",
            current_dir=os.path.dirname(os.path.abspath(__file__)),
        )

    def _generate_client_container_args(
        self, client_src_zip_file_path: str
    ) -> list[str]:
        args = []
        if client_src_zip_file_path:
            args.extend(["--zip", client_src_zip_file_path])
        args.extend(
            self.generate_spark_cmd_args(args=self.args, entrypoint_arg="driver")
        )
        return args

    def _client_image_path_sys_registry(self) -> str:
        return (
            "/snowflake/images/snowflake_images/snowflake-managed-spark:0.0.1-preview"
        )

    def _server_image_path_sys_registry(self) -> str:
        return "/snowflake/images/snowflake_images/spark_connect_for_snowpark_server:0.0.2-preview"

    def _client_image_name_override(self) -> str:
        return "spark_0_0_1:preview"

    def _server_image_name_override(self) -> str:
        return "sas_server_0_0_2:preview"

    def _add_additional_jars_to_classpath(self) -> None:
        # Spark image already contains all needed dependency jars. No need to add more
        pass

    def _use_system_registry(self) -> bool:
        return False

    def _override_args(self) -> None:
        # For classic Spark driver, we do not pass --remote in order to start up
        # full Spark driver context.
        self.args.remote = None



#
# Copyright (c) 2012-2025 Snowflake Computing Inc. All rights reserved.
#

import copy
import os
import sys
from contextlib import contextmanager

import pytest
from snowflake.snowpark_submit.cluster_mode.job_runner import (
    generate_random_name,
    logger,
)
from snowflake.snowpark_submit.cluster_mode.spark_connect.spark_connect_job_runner import (
    SparkConnectJobRunner,
)
from snowflake.snowpark_submit.snowpark_submit import init_args, runner_wrapper

from snowflake.connector.errors import Error
from snowflake.snowpark import Row

from .conftest import DEFAULT_TEST_CONNECTION_NAME

current_dir = os.path.dirname(os.path.abspath(__file__))
RESOURCES_DIR = current_dir + "/resources"

# SAS versions to test against
SAS_VERSIONS = ["latest", "0.20.1"]


@contextmanager
def argv_context(test_args):
    original_argv = sys.argv
    try:
        sys.argv = test_args
        yield
    finally:
        sys.argv = original_argv


@contextmanager
def sas_version_set_via_env_vars(sas_version):
    try:
        os.environ["SNOWFLAKE_SYSTEM_REGISTRY_SERVER_IMAGE_TAG"] = sas_version
        os.environ["SNOWFLAKE_SYSTEM_REGISTRY_CLIENT_IMAGE_TAG"] = sas_version
        yield
    finally:
        os.environ.pop("SNOWFLAKE_SYSTEM_REGISTRY_SERVER_IMAGE_TAG", None)
        os.environ.pop("SNOWFLAKE_SYSTEM_REGISTRY_CLIENT_IMAGE_TAG", None)


class TestSnowparkSubmit:
    @pytest.mark.parametrize("sas_version", SAS_VERSIONS)
    def test_snowpark_submit_cluster_mode_python(
        self,
        spcs_snowflake_config,
        spcs_snowpark_session,
        skip_tests_out_of_aws_us_west_2,
        pyspark_utils_zip,
        sas_version,
    ):
        with sas_version_set_via_env_vars(sas_version):
            pyspark_example_dir = RESOURCES_DIR + "/snowpark-submit-pyspark-example"
            wheel_files_dir = pyspark_example_dir + "/wheel-files"

            workload_name = generate_random_name(
                prefix="snowpark_submit_python_spcs_test_"
            ).upper()
            table_name = "snowpark_submit_python_spcs_test_table"
            spcs_snowpark_session.sql(
                f"create or replace table {table_name} (id int, val float)"
            ).collect()

            integration_name, network_rule_name = self._setup_pypi_external_access(
                spcs_snowpark_session
            )
            if not integration_name:
                pytest.fail("Could not set up external access integration for PyPI")

            test_args = [
                "snowpark-submit",
                "--py-files",
                str(pyspark_utils_zip.resolve()),
                "--requirements-file",
                pyspark_example_dir + "/test-requirements.txt",
                "--wheel-files",
                wheel_files_dir
                + "/access-1.1.9-py3-none-any.whl,"
                + wheel_files_dir
                + "/python_weather-2.1.0-py3-none-any.whl",
                "--external-access-integrations",
                integration_name,
                "--wait-for-completion",
                "--snowflake-workload-name",
                workload_name,
                pyspark_example_dir + "/main.py",
                table_name,
            ]
            self._snowpark_submit_cluster_mode_helper(
                test_args + spcs_snowflake_config,
                spcs_snowpark_session,
                workload_name,
                table_name,
                [Row(ID=1, VAL=2.0), Row(ID=2, VAL=3.0), Row(ID=4, VAL=5.0)],
            )

            spcs_snowpark_session.sql(f"drop table if exists {table_name}").collect()

            self._cleanup_pypi_external_access(
                spcs_snowpark_session, integration_name, network_rule_name
            )

    @pytest.mark.parametrize("sas_version", SAS_VERSIONS)
    def test_snowpark_submit_cluster_mode_scala(
        self,
        spcs_snowflake_config,
        spcs_snowpark_session,
        spcs_scala_test_table,
        skip_tests_out_of_aws_us_west_2,
        sas_version,
    ):
        with sas_version_set_via_env_vars(sas_version):
            scala_example_dir = RESOURCES_DIR + "/snowpark-submit-scala-example"
            workload_name = generate_random_name(
                prefix="snowpark_submit_scala_spcs_test_"
            ).upper()

            test_args = [
                "snowpark-submit",
                "--class",
                "com.example.SnowparkConnectApp",
                "--snowflake-workload-name",
                workload_name,
                scala_example_dir + "/target/original-scala-maven-example-0.1.0.jar",
                spcs_scala_test_table,
            ]
            self._snowpark_submit_cluster_mode_helper(
                test_args + spcs_snowflake_config,
                spcs_snowpark_session,
                workload_name,
                spcs_scala_test_table,
                [
                    Row(NAME="Alice", AGE=29),
                    Row(NAME="Bob", AGE=31),
                    Row(NAME="Catherine", AGE=25),
                ],
            )

    @pytest.mark.parametrize("sas_version", SAS_VERSIONS)
    def test_snowpark_submit_cluster_mode_scala_user_stage(
        self,
        spcs_snowflake_config,
        spcs_snowpark_session,
        spcs_scala_test_table,
        skip_tests_out_of_aws_us_west_2,
        sas_version,
    ):
        with sas_version_set_via_env_vars(sas_version):
            scala_example_dir = RESOURCES_DIR + "/snowpark-submit-scala-example"
            workload_name = generate_random_name(
                prefix="snowpark_submit_scala_spcs_test_"
            ).upper()
            stage_name = "sas_test_spcs_user_stage"
            spcs_snowpark_session.sql(
                f"CREATE OR REPLACE STAGE {stage_name} ENCRYPTION = (TYPE = 'SNOWFLAKE_SSE') DIRECTORY = ( ENABLE = true );"
            ).collect()
            spcs_snowpark_session.sql(
                f"put file://{scala_example_dir}/target/original-scala-maven-example-0.1.0.jar @{stage_name}/snowpark-submit-scala-example/target/ auto_compress=FALSE overwrite=TRUE"
            ).collect()

            test_args = [
                "snowpark-submit",
                "--class",
                "com.example.SnowparkConnectApp",
                "--snowflake-workload-name",
                workload_name,
                "--snowflake-stage",
                f"@{stage_name}",
                f"@{stage_name}/snowpark-submit-scala-example/target/original-scala-maven-example-0.1.0.jar",
                spcs_scala_test_table,
            ]
            self._snowpark_submit_cluster_mode_helper(
                test_args + spcs_snowflake_config,
                spcs_snowpark_session,
                workload_name,
                # table_name,
                spcs_scala_test_table,
                [
                    Row(NAME="Alice", AGE=29),
                    Row(NAME="Bob", AGE=31),
                    Row(NAME="Catherine", AGE=25),
                ],
            )
            spcs_snowpark_session.sql(f"drop stage if exists {stage_name}").collect()

    def _snowpark_submit_cluster_mode_helper(
        self, test_args, snowpark_session, service_name, table_name, expected_res
    ):
        with argv_context(test_args):
            result = runner_wrapper(test_mode=True)
            assert (
                result.exit_code == 0
            ), f"Exit code should be 0, actual: {result.exit_code}"

        status_test_args = [
            "snowpark-submit",
            "--workload-status",
            "--wait-for-completion",
            "--display-logs",
            "--snowflake-workload-name",
            service_name,
            "--snowflake-connection-name",
            DEFAULT_TEST_CONNECTION_NAME,
        ]

        with argv_context(status_test_args):
            status_result = runner_wrapper(test_mode=True)
            assert (
                status_result.exit_code == 0
            ), f"Status check exit code should be 0, actual: {status_result.exit_code}"
            assert (
                status_result.service_status == "DONE"
            ), f"Service status is not DONE: {status_result.service_status}"
            assert (
                status_result.logs is not None
            ), "Logs should not be None when --display-logs is used"
            assert (
                len(status_result.logs) > 1
            ), f"Expected logs to be retrieved, but got empty logs. Status: {status_result.service_status}"

        res = snowpark_session.sql(f"SELECT * FROM {table_name}").collect()
        logger.debug(res)
        assert res == expected_res

        snowpark_session.sql(f"drop service if exists {service_name}").collect()

    def test_execute_service_sql_template(self):
        base_args = init_args(
            [
                "--compute-pool",
                "test_compute_pool",
                "--snowflake-connection-name",
                DEFAULT_TEST_CONNECTION_NAME,
                "--snowflake-workload-name",
                "snowpark_submit_test_workload",
                "workload_test_file.py",
            ]
        )[0]

        def _test_case(args, expected_sql):
            runner = SparkConnectJobRunner(args, None)
            sql = runner.generate_execute_service_sql(
                args.snowflake_workload_name, "<spcs_spec_yaml_file>"
            )
            assert sql == expected_sql

        args = copy.deepcopy(base_args)
        _test_case(
            args,
            """
EXECUTE JOB SERVICE
    IN COMPUTE POOL test_compute_pool
    NAME = snowpark_submit_test_workload
    ASYNC = TRUE
    FROM SPECIFICATION $$
    <spcs_spec_yaml_file>
    $$
    """,
        )

        args = copy.deepcopy(base_args)
        args.external_access_integrations = ["integration1", "INTEGRATION2"]
        _test_case(
            args,
            """
EXECUTE JOB SERVICE
    IN COMPUTE POOL test_compute_pool
    NAME = snowpark_submit_test_workload
    ASYNC = TRUE
    EXTERNAL_ACCESS_INTEGRATIONS = (integration1, INTEGRATION2)
    FROM SPECIFICATION $$
    <spcs_spec_yaml_file>
    $$
    """,
        )

    @pytest.mark.parametrize("sas_version", SAS_VERSIONS)
    def test_snowpark_submit_cluster_mode_python_without_connection_file(
        self,
        spcs_snowpark_session,
        skip_tests_out_of_aws_us_west_2,
        pyspark_utils_zip,
        sas_version,
    ):
        with sas_version_set_via_env_vars(sas_version):
            from snowflake.connector.config_manager import CONFIG_MANAGER

            connection_config = CONFIG_MANAGER["connections"][
                DEFAULT_TEST_CONNECTION_NAME
            ]

            pyspark_example_dir = RESOURCES_DIR + "/snowpark-submit-pyspark-example"
            wheel_files_dir = pyspark_example_dir + "/wheel-files"

            workload_name = generate_random_name(
                prefix="snowpark_submit_py_spcs_cli_"
            ).upper()
            table_name = "snowpark_submit_python_spcs_test_table_cli"
            spcs_snowpark_session.sql(
                f"create or replace table {table_name} (id int, val float)"
            ).collect()

            integration_name, network_rule_name = self._setup_pypi_external_access(
                spcs_snowpark_session
            )
            if not integration_name:
                pytest.fail("Could not set up external access integration for PyPI")
            # use explicit connection parameters
            test_args = [
                "snowpark-submit",
                "--py-files",
                str(pyspark_utils_zip.resolve()),
                "--requirements-file",
                pyspark_example_dir + "/test-requirements.txt",
                "--wheel-files",
                wheel_files_dir
                + "/access-1.1.9-py3-none-any.whl,"
                + wheel_files_dir
                + "/python_weather-2.1.0-py3-none-any.whl",
                "--external-access-integrations",
                integration_name,
                "--wait-for-completion",
                "--snowflake-workload-name",
                workload_name,
                "--account",
                connection_config.get("account"),
                "--host",
                connection_config.get("host"),
                "--user",
                connection_config.get("user"),
                "--password",
                connection_config.get("password"),
                "--compute-pool",
                connection_config.get("compute_pool"),
            ]

            # optional
            if "database" in connection_config:
                test_args.extend(["--database", connection_config["database"]])
            if "warehouse" in connection_config:
                test_args.extend(["--warehouse", connection_config["warehouse"]])
            if "schema" in connection_config:
                test_args.extend(["--schema", connection_config["schema"]])
            if "role" in connection_config:
                test_args.extend(["--role", connection_config["role"]])

            test_args.extend(
                [
                    "--snowflake-log-level",
                    "NONE",
                    pyspark_example_dir + "/main.py",
                    table_name,
                ]
            )

            self._snowpark_submit_cluster_mode_helper(
                test_args,
                spcs_snowpark_session,
                workload_name,
                table_name,
                [Row(ID=1, VAL=2.0), Row(ID=2, VAL=3.0), Row(ID=4, VAL=5.0)],
            )

            spcs_snowpark_session.sql(f"drop table if exists {table_name}").collect()

            self._cleanup_pypi_external_access(
                spcs_snowpark_session, integration_name, network_rule_name
            )

    def test_snowpark_submit_missing_required_params(self):  # UNIT TEST
        pyspark_example_dir = RESOURCES_DIR + "/snowpark-submit-pyspark-example"

        def _test_missing_param(test_args, expected_missing_params):
            with argv_context(test_args):
                with pytest.raises(Error) as exc_info:
                    runner_wrapper(test_mode=True)
                error_message = str(exc_info.value)
                assert "Missing required connection parameters" in error_message
                for param in expected_missing_params:
                    assert param in error_message

        base_args = [
            "snowpark-submit",
            "--snowflake-workload-name",
            "test_workload",
            pyspark_example_dir + "/main.py",
        ]

        # test missing --account
        test_args = base_args.copy()
        test_args.extend(
            [
                "--host",
                "test.snowflakecomputing.com",
                "--user",
                "test_user",
                "--password",
                "test_password",
                "--compute-pool",
                "test_pool",
            ]
        )
        _test_missing_param(test_args, ["--account"])

        # test missing --host
        test_args = base_args.copy()
        test_args.extend(
            [
                "--account",
                "test_account",
                "--user",
                "test_user",
                "--password",
                "test_password",
                "--compute-pool",
                "test_pool",
            ]
        )
        _test_missing_param(test_args, ["--host"])

        # Test case missing --compute-pool
        test_args = base_args.copy()
        test_args.extend(
            [
                "--account",
                "test_account",
                "--host",
                "test.snowflakecomputing.com",
                "--user",
                "test_user",
                "--password",
                "test_password",
            ]
        )
        _test_missing_param(test_args, ["--compute-pool"])

        # Test when missing multiple parameters
        test_args = base_args.copy()
        test_args.extend(
            [
                "--user",
                "test_user",
                "--password",
                "test_password",
            ]
        )
        _test_missing_param(test_args, ["--account", "--host", "--compute-pool"])

    def test_snowpark_submit_cli_only_mode_session_config(self):
        # Tests that session config is created correctly when using CLI args only

        test_account = "test_account"
        test_host = "test.snowflakecomputing.com"
        test_user = "test_user"
        test_password = "test_password"
        test_compute_pool = "test_compute_pool"

        test_args = [
            "--account",
            test_account,
            "--host",
            test_host,
            "--user",
            test_user,
            "--password",
            test_password,
            "--compute-pool",
            test_compute_pool,
            "--snowflake-workload-name",
            "test_workload",
            "dummy_file.py",
        ]

        args, _ = init_args(test_args)

        assert args.snowflake_connection_name is None

        runner = SparkConnectJobRunner(args, None)

        expected_config = {
            "account": test_account,
            "host": test_host,
            "user": test_user,
            "password": test_password,
            "protocol": "https",
            "port": 443,
        }

        # Assert that session config was built correctly from CLI args, not from a connection file
        assert runner.custom_session_configs is not None
        for key, expected_value in expected_config.items():
            assert runner.custom_session_configs.get(key) == expected_value

        assert "connection_name" not in runner.custom_session_configs
        assert runner.args.compute_pool == test_compute_pool

    def test_snowpark_submit_cli_args_override_connection_file(self):  # unit test
        # provides an invalid password to override the one in the connection file
        # asserts that the connection file password is used instead of the CLI provided one

        pyspark_example_dir = RESOURCES_DIR + "/snowpark-submit-pyspark-example"

        test_args = [
            "snowpark-submit",
            "--snowflake-workload-name",
            "test_override_workload",
            "--snowflake-connection-name",
            DEFAULT_TEST_CONNECTION_NAME,
            "--password",
            "INVALID_PASSWORD_12345",
            pyspark_example_dir + "/main.py",
            "dummy_table",
        ]

        from snowflake.connector.errors import DatabaseError

        with argv_context(test_args):
            with pytest.raises(DatabaseError) as exc_info:
                runner_wrapper(test_mode=True)

            error_msg = str(exc_info.value)
            assert (
                "250001" in error_msg or "08001" in error_msg
            ), f"Expected authentication error code but got: {error_msg}"
            assert (
                "Incorrect username or password was specified" in error_msg
            ), f"Expected password error but got: {error_msg}"

    def test_snowpark_submit_validation_no_operation_specified(self):  # UNIT TEST
        """Test that validation fails when no operation is specified"""
        test_args = [
            "snowpark-submit",
            "--snowflake-connection-name",
            DEFAULT_TEST_CONNECTION_NAME,
        ]

        with argv_context(test_args):
            result = runner_wrapper(test_mode=True)
            assert result.exit_code == 1
            assert "You must specify exactly one operation at a time" in result.error

    def test_snowpark_submit_validation_multiple_operations_specified(
        self,
    ):  # UNIT TEST
        """Test that validation fails when multiple operations are specified"""
        pyspark_example_dir = RESOURCES_DIR + "/snowpark-submit-pyspark-example"

        test_args = [
            "snowpark-submit",
            "--snowflake-connection-name",
            DEFAULT_TEST_CONNECTION_NAME,
            "--workload-status",
            "--kill-workload",
            pyspark_example_dir + "/main.py",
        ]

        with argv_context(test_args):
            result = runner_wrapper(test_mode=True)
            assert result.exit_code == 1
            assert "You must specify exactly one operation at a time" in result.error

    def test_end_workload_missing_workload_name(self):  # UNIT TEST
        """Test that end_workload fails when workload name is not provided"""
        test_args = [
            "snowpark-submit",
            "--snowflake-connection-name",
            DEFAULT_TEST_CONNECTION_NAME,
            "--kill-workload",
        ]

        with argv_context(test_args):
            result = runner_wrapper(test_mode=True)
            assert result.exit_code == 1
            assert "Missing mandatory option --snowflake-workload-name" in result.error

    def test_end_workload_success(self):  # INTEGRATION TEST
        """Test complete workload lifecycle: create -> check -> kill -> verify"""

        workload_name = generate_random_name(
            prefix="snowpark_submit_lifecycle_test_"
        ).upper()

        pyspark_example_dir = RESOURCES_DIR + "/snowpark-submit-pyspark-example"

        try:
            # Step 1: Submit a workload
            create_args = [
                "snowpark-submit",
                "--snowflake-connection-name",
                DEFAULT_TEST_CONNECTION_NAME,
                "--snowflake-workload-name",
                workload_name,
                pyspark_example_dir + "/main.py",
                "dummy_table",
            ]

            with argv_context(create_args):
                create_result = runner_wrapper(test_mode=True)
                assert (
                    create_result.exit_code == 0
                ), f"Failed to create workload: {create_result.error}"

            # Step 2: Check if workload exists
            status_args = [
                "snowpark-submit",
                "--snowflake-connection-name",
                DEFAULT_TEST_CONNECTION_NAME,
                "--workload-status",
                "--snowflake-workload-name",
                workload_name,
            ]

            with argv_context(status_args):
                status_result = runner_wrapper(test_mode=True)
                assert (
                    status_result.exit_code == 0
                ), f"Failed to get workload status: {status_result.error}"
                assert status_result.workload_name == workload_name
                assert (
                    status_result.service_status is not None
                )  # Should have some status

            # Step 3: Kill the workload
            kill_args = [
                "snowpark-submit",
                "--snowflake-connection-name",
                DEFAULT_TEST_CONNECTION_NAME,
                "--kill-workload",
                "--snowflake-workload-name",
                workload_name,
            ]

            with argv_context(kill_args):
                kill_result = runner_wrapper(test_mode=True)
                assert (
                    kill_result.exit_code == 0
                ), f"Failed to kill workload: {kill_result.error}"

            # Step 4: Verify workload is gone/terminated
            final_status_args = [
                "snowpark-submit",
                "--snowflake-connection-name",
                DEFAULT_TEST_CONNECTION_NAME,
                "--workload-status",
                "--snowflake-workload-name",
                workload_name,
            ]

            with argv_context(final_status_args):
                final_result = runner_wrapper(test_mode=True)
                # Should either fail (service doesn't exist) or show terminated status
                if final_result.exit_code == 0:
                    # If it still exists, it should be in a terminated state
                    assert (
                        final_result.terminated is True
                    ), f"Workload should be terminated but status is: {final_result.service_status}"
                else:
                    # Or it could be completely gone, which is also acceptable
                    assert "does not exist" in final_result.error

        except Exception as e:
            # Cleanup: Try to drop the service if test fails
            try:
                cleanup_args = [
                    "snowpark-submit",
                    "--snowflake-connection-name",
                    DEFAULT_TEST_CONNECTION_NAME,
                    "--kill-workload",
                    "--snowflake-workload-name",
                    workload_name,
                ]
                with argv_context(cleanup_args):
                    runner_wrapper(test_mode=True)
            except Exception:
                pass  # Ignore cleanup failures
            raise e

    @pytest.mark.parametrize("sas_version", SAS_VERSIONS)
    def test_snowpark_connect_version_param_python(
        self,
        spcs_snowflake_config,
        spcs_snowpark_session,
        skip_tests_out_of_aws_us_west_2,
        pyspark_utils_zip,
        sas_version,
    ):
        with sas_version_set_via_env_vars("invalid"):

            pyspark_example_dir = RESOURCES_DIR + "/snowpark-submit-pyspark-example"
            wheel_files_dir = pyspark_example_dir + "/wheel-files"

            workload_name = generate_random_name(
                prefix="snowpark_submit_version_test_"
            ).upper()
            table_name = "snowpark_submit_version_test_table"

            spcs_snowpark_session.sql(
                f"create or replace table {table_name} (id int, val float)"
            ).collect()

            integration_name, network_rule_name = self._setup_pypi_external_access(
                spcs_snowpark_session
            )
            if not integration_name:
                pytest.fail("Could not set up external access integration for PyPI")

            test_args = [
                "snowpark-submit",
                "--py-files",
                str(pyspark_utils_zip.resolve()),
                "--requirements-file",
                pyspark_example_dir + "/test-requirements.txt",
                "--wheel-files",
                wheel_files_dir
                + "/access-1.1.9-py3-none-any.whl,"
                + wheel_files_dir
                + "/python_weather-2.1.0-py3-none-any.whl",
                "--external-access-integrations",
                integration_name,
                "--wait-for-completion",
                "--snowflake-workload-name",
                workload_name,
                "--snowpark-connect-version",
                sas_version,
                pyspark_example_dir + "/main.py",
                table_name,
            ]

            self._snowpark_submit_cluster_mode_helper(
                test_args + spcs_snowflake_config,
                spcs_snowpark_session,
                workload_name,
                table_name,
                [Row(ID=1, VAL=2.0), Row(ID=2, VAL=3.0), Row(ID=4, VAL=5.0)],
            )

            spcs_snowpark_session.sql(f"drop table if exists {table_name}").collect()

            self._cleanup_pypi_external_access(
                spcs_snowpark_session, integration_name, network_rule_name
            )

    @pytest.mark.parametrize("sas_version", SAS_VERSIONS)
    def test_snowpark_connect_version_param_scala(
        self,
        spcs_snowflake_config,
        spcs_snowpark_session,
        spcs_scala_test_table,
        skip_tests_out_of_aws_us_west_2,
        sas_version,
    ):
        with sas_version_set_via_env_vars("invalid"):

            scala_example_dir = RESOURCES_DIR + "/snowpark-submit-scala-example"
            workload_name = generate_random_name(
                prefix="snowpark_submit_scala_spcs_test_"
            ).upper()

            test_args = [
                "snowpark-submit",
                "--class",
                "com.example.SnowparkConnectApp",
                "--snowflake-workload-name",
                workload_name,
                "--snowpark-connect-version",
                sas_version,
                scala_example_dir + "/target/original-scala-maven-example-0.1.0.jar",
                spcs_scala_test_table,
            ]
            self._snowpark_submit_cluster_mode_helper(
                test_args + spcs_snowflake_config,
                spcs_snowpark_session,
                workload_name,
                spcs_scala_test_table,
                [
                    Row(NAME="Alice", AGE=29),
                    Row(NAME="Bob", AGE=31),
                    Row(NAME="Catherine", AGE=25),
                ],
            )

    def _setup_pypi_external_access(self, session):
        uid = generate_random_name(prefix="test_pypi_")
        network_rule_name = f"pypi_network_rule_{uid}".upper()
        integration_name = f"pypi_access_integration_{uid}".upper()

        current_role = session.get_current_role()

        try:
            session.sql(
                f"""
                CREATE OR REPLACE NETWORK RULE {network_rule_name}
                  MODE = EGRESS
                  TYPE = HOST_PORT
                  VALUE_LIST = ('pypi.org', 'files.pythonhosted.org')
                """
            ).collect()
            logger.debug(
                f"Created network rule '{network_rule_name}' with role {current_role}"
            )

            session.sql("USE ROLE SYSADMIN").collect()

            session.sql(
                f"""
                CREATE OR REPLACE EXTERNAL ACCESS INTEGRATION {integration_name}
                  ALLOWED_NETWORK_RULES = ({network_rule_name})
                  ENABLED = true
                  COMMENT = 'External access integration for PyPI package installation (test)'
                """
            ).collect()
            logger.debug(
                f"Created external access integration '{integration_name}' with SYSADMIN role"
            )

            session.sql(
                f"GRANT USAGE ON INTEGRATION {integration_name} TO ROLE {current_role}"
            ).collect()
            logger.debug(f"Granted usage on integration to role {current_role}")

            session.sql(f"USE ROLE {current_role}").collect()

            return integration_name, network_rule_name
        except Exception as e:
            logger.error(f"Failed to set up external access integration: {e}")
            raise Exception(f"Failed to set up external access integration: {e}") from e

    def _cleanup_pypi_external_access(
        self, session, integration_name, network_rule_name
    ):
        try:
            if integration_name:
                session.sql(f"DROP INTEGRATION IF EXISTS {integration_name}").collect()
                logger.debug(f"Dropped integration '{integration_name}'")
            if network_rule_name:
                session.sql(
                    f"DROP NETWORK RULE IF EXISTS {network_rule_name}"
                ).collect()
                logger.debug(f"Dropped network rule '{network_rule_name}'")
        except Exception as e:
            logger.warning(f"Failed to clean up external access resources: {e}")



#
# Copyright (c) 2012-2025 Snowflake Computing Inc. All rights reserved.
#

import os
import random
import string
import sys
from io import StringIO
from typing import Optional

from airflow.models import BaseOperator
from airflow.utils.context import Context
from snowflake.snowpark_submit.snowpark_submit import runner_wrapper


def generate_service_name(
    prefix: str = "SNOWPARK_SUBMIT_AIRFLOW_OPERATOR_", length: int = 27
) -> str:
    choices = string.ascii_uppercase + string.digits
    random_id = "".join(random.choice(choices) for _ in range(length))
    return f"{prefix}{random_id}"


class SnowparkSubmitOperator(BaseOperator):
    """
    Apache Airflow operator for executing snowpark-submit commands.

    :param file: Path to the file to execute (Python or JAR file).
    :param main_class: Main class name for Java/Scala applications (required for JAR files)
    :param application_args: List of arguments to pass to the application
    :param connections_config: Dictionary containing Snowflake connection parameters
    :param service_name: Custom service name (auto-generated if not provided)
    :param py_files: Comma-separated list of Python dependencies (Python only)
    :param jars: Comma-separated list of JAR dependencies (Java/Scala only)
    :param name: A name of your application.
    :param comment: Comment/description for the workload
    :param snowflake_log_level: Log level for Snowflake events ('INFO', 'ERROR', 'NONE')
    :param wait_for_completion: Run workload in blocking mode and wait for completion (default: False)
    :param properties_file: Path to a file from which to load extra Spark properties
    :param fail_on_error: Whether to fail the Airflow task if the Spark job failed (only applies when wait_for_completion=True)

    Ex:
        task = SnowparkSubmitOperator(
            task_id='my_python_job',
            file='/path/to/script.py',
            connections_config={
                'account': snowflake_conn.extra_dejson.get('account'),
                'host': snowflake_conn.host,
                'user': snowflake_conn.login,
                'password': snowflake_conn.password,
                'role': snowflake_conn.extra_dejson.get('role'),
                'warehouse': snowflake_conn.extra_dejson.get('warehouse'),
                'database': snowflake_conn.schema,
                'schema': snowflake_conn.extra_dejson.get('schema'),
                'compute_pool': snowflake_conn.extra_dejson.get('compute_pool'),
            },
            dag=dag
        )

    """

    template_fields = [
        "file",
        "main_class",
        "application_args",
        "service_name",
        "py_files",
        "jars",
        "name",
        "connections_config",
        "comment",
        "snowflake_log_level",
        "wait_for_completion",
        "properties_file",
        "fail_on_error",
    ]

    def __init__(
        self,
        connections_config: dict,
        file: Optional[str] = None,
        main_class: Optional[str] = None,
        application_args: Optional[list] = None,
        service_name: Optional[str] = None,
        py_files: Optional[str] = None,
        jars: Optional[str] = None,
        name: Optional[str] = None,
        comment: Optional[str] = None,
        snowflake_log_level: Optional[str] = None,
        wait_for_completion: bool = False,
        properties_file: Optional[str] = None,
        fail_on_error: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.file = file
        self.main_class = main_class
        self.application_args = application_args or []
        self.connections_config = connections_config
        self.service_name = service_name
        self.py_files = py_files
        self.jars = jars
        self.name = name
        self.comment = comment
        self.snowflake_log_level = snowflake_log_level
        self.wait_for_completion = wait_for_completion
        self.properties_file = properties_file
        self.fail_on_error = fail_on_error

    def _validate_required_config(self):
        required_fields = ["account", "host", "compute_pool"]

        missing_fields = []
        for field in required_fields:
            if (
                field not in self.connections_config
                or not self.connections_config[field]
            ):
                missing_fields.append(field)
        if missing_fields:
            raise ValueError(
                f"Missing required Snowflake configuration parameters: {', '.join(missing_fields)}"
            )

    def _validate_file(self):
        if not self.file:
            raise ValueError("file parameter must be specified")

        if not os.path.exists(self.file):
            raise FileNotFoundError(f"File not found: {self.file}")

        file_ext = os.path.splitext(self.file)[1].lower()
        if file_ext == ".py":
            self.log.info(f"Detected Python file: {self.file}")
        elif file_ext == ".jar":
            self.log.info(f"Detected JAR file: {self.file}")
            if not self.main_class:
                raise ValueError("main_class is required for JAR files")
        else:
            raise ValueError(
                f"Unsupported file type: {file_ext}. Only .py and .jar files are supported"
            )

    def _build_command(self) -> list:
        cmd = ["snowpark-submit"]

        if self.main_class:
            cmd.extend(["--class", self.main_class])
        for key, value in self.connections_config.items():
            if value is not None:
                cli_key = key.replace("_", "-")
                cmd.extend([f"--{cli_key}", str(value)])

        if self.name:
            cmd.extend(["--name", self.name])
        if self.comment:
            cmd.extend(["--comment", self.comment])
        if self.snowflake_log_level:
            cmd.extend(["--snowflake-log-level", self.snowflake_log_level])
        if self.jars:
            cmd.extend(["--jars", self.jars])
        if self.py_files:
            cmd.extend(["--py-files", self.py_files])

        if self.wait_for_completion:
            cmd.append("--wait-for-completion")
            self.log.info("Running in blocking mode will wait for completion")
        if self.properties_file:
            cmd.extend(["--properties-file", self.properties_file])
            self.log.info(f"Using properties file: {self.properties_file}")

        if not self.service_name:
            self.service_name = generate_service_name()
            self.log.info(f"generated service name: {self.service_name}")
        else:
            self.log.info(f"Using service name: {self.service_name}")

        cmd.append(f"--snowflake-workload-name={self.service_name}")

        if self.file:
            cmd.append(self.file)
        else:
            raise ValueError("file parameter must be specified")

        if self.application_args:
            cmd.extend(self.application_args)

        return cmd

    def execute(self, context: Context):
        self._validate_required_config()
        self._validate_file()

        if self.fail_on_error and not self.wait_for_completion:
            self.log.warning(
                "fail_on_error parameter is ignored when wait_for_completion=False. "
                "fail_on_error only applies to blocking mode."
            )

        if self.properties_file:
            if not os.path.exists(self.properties_file):
                raise FileNotFoundError(
                    f"Properties file not found: {self.properties_file}"
                )

        cmd = self._build_command()

        self.log.info(f"Executing snowpark-submit command: {' '.join(cmd)}")

        original_stdout = sys.stdout
        original_stderr = sys.stderr
        original_argv = sys.argv

        captured_stdout = StringIO()
        captured_stderr = StringIO()

        try:
            sys.stdout = captured_stdout
            sys.stderr = captured_stderr
            sys.argv = cmd

            result = runner_wrapper(test_mode=True)
            stdout_content = captured_stdout.getvalue()
            stderr_content = captured_stderr.getvalue()

            self.log.info(f"runner_wrapper returned with exit code: {result.exit_code}")
            context["ti"].xcom_push(key="service_name", value=self.service_name)  # type: ignore
            context["ti"].xcom_push(key="return_code", value=result.exit_code)  # type: ignore
            context["ti"].xcom_push(key="stdout", value=stdout_content)  # type: ignore
            context["ti"].xcom_push(key="stderr", value=stderr_content)  # type: ignore

            job_failed = False

            # alwas fail if snowpark-submit command itself failed
            if result.exit_code != 0:
                job_failed = True
                self.log.error(
                    f"snowpark-submit command failed with exit code: {result.exit_code}"
                )

            # if submission succeeded but we're in blocking mode, check for Spark job failure
            elif self.wait_for_completion and self.fail_on_error:
                # check if the Spark job itself failed (even though submission succeeded)
                if "Workload Status: FAILED" in stdout_content:
                    job_failed = True
                    self.log.error("Spark job failed - workload status: FAILED")

            if not job_failed:
                if result.exit_code == 0:
                    self.log.info(
                        f"snowpark-submit completed successfully with exit code: {result.exit_code}"
                    )
                    self.log.info(f"Job service created with name: {self.service_name}")

            if job_failed:
                if result.exit_code != 0:
                    raise Exception(
                        f"snowpark-submit command failed with exit code: {result.exit_code}"
                    )
                else:
                    raise Exception("Spark job failed with status: FAILED")

            return {
                "service_name": self.service_name,
                "return_code": result.exit_code,
                "stdout": stdout_content,
                "stderr": stderr_content,
            }

        except Exception as e:
            self.log.error(f"snowpark-submit failed: {e}")
            raise
        finally:
            sys.stdout = original_stdout
            sys.stderr = original_stderr
            sys.argv = original_argv



#
# Copyright (c) 2012-2025 Snowflake Computing Inc. All rights reserved.
#

import os
import time
from unittest.mock import Mock

import pytest
from airflow.utils.context import Context
from snowflake.operators.snowpark_submit.snowpark_submit_operator import (
    SnowparkSubmitOperator,
)
from snowflake.operators.snowpark_submit.snowpark_submit_status_operator import (
    SnowparkSubmitStatusOperator,
)
from snowflake.snowpark_submit.cluster_mode.job_runner import logger

from snowflake.connector.config_manager import CONFIG_MANAGER

from .conftest import DEFAULT_TEST_CONNECTION_NAME


class TestSnowparkOperatorsIntegration:
    @classmethod
    def setup_class(cls):
        cls.tests_path = os.path.dirname(os.path.abspath(__file__))
        cls.project_root = os.path.dirname(cls.tests_path)

        snowpark_config = CONFIG_MANAGER["connections"][DEFAULT_TEST_CONNECTION_NAME]
        cls.connection_config = {
            "account": snowpark_config.get("account"),
            "host": snowpark_config.get("host"),
            "user": snowpark_config.get("user"),
            "password": snowpark_config.get("password"),
            "role": snowpark_config.get("role"),
            "warehouse": snowpark_config.get("warehouse"),
            "database": snowpark_config.get("database"),
            "schema": snowpark_config.get("schema"),
            "compute_pool": snowpark_config.get("compute_pool"),
        }

        cls.mock_ti = Mock()
        cls.mock_ti.xcom_push = Mock()
        cls.mock_context = Context()
        cls.mock_context["ti"] = cls.mock_ti

    def test_submit_and_check_status(self):
        # Submit a job and check its status
        submit_operator = SnowparkSubmitOperator(
            task_id="test_submit_job",
            file=os.path.join(self.tests_path, "resources/pyspark-test.py"),
            connections_config=self.connection_config,
            wait_for_completion=False,
        )

        result = submit_operator.execute(self.mock_context)

        assert result is not None
        assert result["return_code"] == 0
        assert result["service_name"] is not None
        assert result["service_name"].startswith("SNOWPARK_SUBMIT_AIRFLOW_OPERATOR_")

        stdout_content = result["stdout"]
        service_name = result["service_name"]

        job_msg = "has been submitted and running asynchorously in SPCS"
        assert (
            job_msg in stdout_content
        ), f"Job submission to SPCS not confirmed for service {service_name}"

        status_operator = SnowparkSubmitStatusOperator(
            task_id="test_check_status",
            snowflake_workload_name=service_name,
            connections_config=self.connection_config,
            display_logs=False,
        )

        status_result = status_operator.execute(self.mock_context)

        assert status_result is not None
        assert (status_result.workload_name or service_name) == service_name
        assert status_result.service_status in ["PENDING", "RUNNING", "DONE"]
        assert status_result.workload_status in ["PENDING", "RUNNING", "DONE"]
        assert isinstance(status_result.created_on, str) and status_result.created_on

    @pytest.mark.skip(
        reason="Requires SPCS testing infra. Temporarily skipped for to avoid long merge gate time."
    )
    def test_submit_wait_and_check_status(self):
        # Submit a job with wait_for_completion=True and check status with display_logs=True
        submit_operator = SnowparkSubmitOperator(
            task_id="test_submit_wait",
            file=os.path.join(self.tests_path, "resources/pyspark-test.py"),
            connections_config=self.connection_config,
            wait_for_completion=True,
        )

        submit_result = submit_operator.execute(self.mock_context)

        assert submit_result is not None
        assert submit_result["return_code"] == 0
        assert "Service Status: DONE" in submit_result["stdout"]

        service_name = submit_result["service_name"]
        status_operator = SnowparkSubmitStatusOperator(
            task_id="test_check_logs",
            snowflake_workload_name=service_name,
            connections_config=self.connection_config,
            display_logs=True,
            log_fetch_delay=5,
        )

        status_result = status_operator.execute(self.mock_context)

        assert status_result is not None
        assert (status_result.workload_name or service_name) == service_name
        assert status_result.service_status == "DONE"
        assert status_result.workload_status == "DONE"
        assert isinstance(status_result.created_on, str) and status_result.created_on
        assert (
            isinstance(status_result.terminated_at, str) and status_result.terminated_at
        )

    @pytest.mark.skip(
        reason="Requires SPCS testing infra. Temporarily skipped for to avoid long merge gate time. Can be enabled in the future with more testing resources."
    )
    def test_submit_failing_job_and_check_status(self):
        # Submit a failing job and verify FAILED status
        submit_operator = SnowparkSubmitOperator(
            task_id="test_submit_failing",
            file=os.path.join(self.tests_path, "resources/pyspark-test-failing.py"),
            connections_config=self.connection_config,
            wait_for_completion=True,
        )

        submit_result = submit_operator.execute(self.mock_context)

        assert submit_result is not None
        assert submit_result["return_code"] == 0
        assert "Service Status: FAILED" in submit_result["stdout"]

        service_name = submit_result["service_name"]
        logger.debug(f"Failing job completed with service name: {service_name}")

        status_operator = SnowparkSubmitStatusOperator(
            task_id="test_check_failed_status",
            snowflake_workload_name=service_name,
            connections_config=self.connection_config,
            display_logs=True,
            log_fetch_delay=5,
        )

        status_result = status_operator.execute(self.mock_context)

        assert status_result is not None
        assert (status_result.workload_name or service_name) == service_name
        assert status_result.service_status == "FAILED"
        assert status_result.workload_status == "FAILED"
        assert isinstance(status_result.created_on, str) and status_result.created_on
        assert (
            isinstance(status_result.terminated_at, str) and status_result.terminated_at
        )

        assert (
            status_result.job_exit_code != 0
        ), f"Expected non-zero exit code but got: {status_result.job_exit_code}"

    @pytest.mark.skip(
        reason="Log retrieval can be flaky. This test should only be run locally."
    )
    def test_submit_wait_and_check_logs(self):
        # Submit a job with wait_for_completion=True, check status, and check logs
        submit_operator = SnowparkSubmitOperator(
            task_id="test_submit_wait",
            file=os.path.join(self.tests_path, "resources/pyspark-test.py"),
            connections_config=self.connection_config,
            wait_for_completion=True,
        )

        submit_result = submit_operator.execute(self.mock_context)

        assert submit_result is not None
        assert submit_result["return_code"] == 0
        assert "Service Status: DONE" in submit_result["stdout"]

        service_name = submit_result["service_name"]
        status_operator = SnowparkSubmitStatusOperator(
            task_id="test_check_logs",
            snowflake_workload_name=service_name,
            connections_config=self.connection_config,
            display_logs=True,
            log_fetch_delay=15,
        )

        status_result = status_operator.execute(self.mock_context)

        assert status_result is not None
        assert (status_result.workload_name or service_name) == service_name
        assert status_result.service_status == "DONE"
        assert status_result.workload_status == "DONE"
        assert isinstance(status_result.created_on, str) and status_result.created_on
        assert (
            isinstance(status_result.terminated_at, str) and status_result.terminated_at
        )

        max_retries = 8
        base_delay = 2
        # wait for logs to be available
        for attempt in range(max_retries):
            if attempt > 0:  # dont need to re-fetch immediately on first attempt
                status_result = status_operator.execute(self.mock_context)

            logger.debug(
                f"Attempt {attempt + 1}: logs count = {len(status_result.logs)}"
            )

            if any(
                "-------PYSPARK TEST JOB SUCCEEDED--------" in log
                for log in status_result.logs
            ):
                break

            if attempt < max_retries - 1:
                # Exponential backoff: 2, 4, 8, 16, 32, ... seconds (capped at 30s)
                delay = min(base_delay * (2**attempt), 30)
                logger.debug(
                    f"Success message not found, waiting {delay} seconds before retry..."
                )
                time.sleep(delay)
            else:
                logger.debug("Last few log entries:")
                for i, log in enumerate(status_result.logs[-3:]):
                    logger.debug(f"  {i}: {log}")
                raise AssertionError(
                    f"Expected log entry not found in logs after {max_retries} attempts"
                )

    @pytest.mark.skip(
        reason="Log retrieval can be flaky. This test should only be run locally."
    )
    def test_submit_failing_job_and_check_logs(self):
        # Submit a failing job, verify FAILED status, check logs for relevant error
        submit_operator = SnowparkSubmitOperator(
            task_id="test_submit_failing",
            file=os.path.join(self.tests_path, "resources/pyspark-test-failing.py"),
            connections_config=self.connection_config,
            wait_for_completion=True,
        )

        submit_result = submit_operator.execute(self.mock_context)

        assert submit_result is not None
        assert submit_result["return_code"] == 0
        assert "Service Status: FAILED" in submit_result["stdout"]

        service_name = submit_result["service_name"]
        logger.debug(f"Failing job completed with service name: {service_name}")

        status_operator = SnowparkSubmitStatusOperator(
            task_id="test_check_failed_status",
            snowflake_workload_name=service_name,
            connections_config=self.connection_config,
            display_logs=True,
            log_fetch_delay=15,
        )

        status_result = status_operator.execute(self.mock_context)

        assert status_result is not None
        assert (status_result.workload_name or service_name) == service_name
        assert status_result.service_status == "FAILED"
        assert status_result.workload_status == "FAILED"
        assert isinstance(status_result.created_on, str) and status_result.created_on
        assert (
            isinstance(status_result.terminated_at, str) and status_result.terminated_at
        )

        assert (
            status_result.job_exit_code != 0
        ), f"Expected non-zero exit code but got: {status_result.job_exit_code}"

        max_retries = 8
        base_delay = 2
        # wait for logs to be available
        for attempt in range(max_retries):
            if attempt > 0:
                status_result = status_operator.execute(self.mock_context)

            logger.debug(
                f"Attempt {attempt + 1}: logs count = {len(status_result.logs)}"
            )

            if any("thisShouldFail" in log for log in status_result.logs):
                logger.debug("Error indicators found in logs (as expected)")
                break

            if attempt < max_retries - 1:
                # Exponential backoff: 2, 4, 8, 16, 32, ... seconds (capped at 30s)
                delay = min(base_delay * (2**attempt), 30)
                logger.debug(
                    f"Error indicators not found, waiting {delay} seconds before retry..."
                )
                time.sleep(delay)
            else:
                logger.debug("Last few log entries:")
                for i, log in enumerate(status_result.logs[-3:]):
                    logger.debug(f"  {i}: {log}")
                raise AssertionError(
                    f"Expected error indicators not found in logs after {max_retries} attempts"
                )

    @pytest.mark.skip(
        reason="Requires SPCS testing infra. Temporarily skipped for to avoid long merge gate time. Can be enabled in the future with more testing resources."
    )
    def test_submit_async_then_wait_for_completion(self):
        # submit async, then use status operator with wait_for_completion=True
        submit_operator = SnowparkSubmitOperator(
            task_id="test_submit_async",
            file=os.path.join(self.tests_path, "resources/pyspark-test.py"),
            connections_config=self.connection_config,
            wait_for_completion=False,
        )

        submit_result = submit_operator.execute(self.mock_context)

        assert submit_result is not None
        assert submit_result["return_code"] == 0
        assert submit_result["service_name"] is not None
        assert submit_result["service_name"].startswith(
            "SNOWPARK_SUBMIT_AIRFLOW_OPERATOR_"
        )

        service_name = submit_result["service_name"]
        logger.debug(f"Submitted async job with service name: {service_name}")

        status_operator = SnowparkSubmitStatusOperator(
            task_id="test_wait_for_existing",
            snowflake_workload_name=service_name,
            connections_config=self.connection_config,
            wait_for_completion=True,
            display_logs=False,
        )

        status_result = status_operator.execute(self.mock_context)

        assert status_result is not None
        assert (status_result.workload_name or service_name) == service_name
        assert status_result.service_status == "DONE"
        assert status_result.workload_status == "DONE"
        assert isinstance(status_result.created_on, str) and status_result.created_on
        assert (
            isinstance(status_result.terminated_at, str) and status_result.terminated_at
        ), "Expected terminated_at to be set for completed workload"
